/*
 * Copyright (C) 2004  Martin Hedler
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Last change:
 * 2004-02-17 by Martin Hedler (matin@hpfsc.de) 
 * 
 */
package ga;

import java.io.Serializable;

/**
 * Stellt ein Chromosom zur Verwaltung von boolean Werten zur Verf&uuml;gung.
 *
 * @author Martin Hedler
 * @version 0.7
 */
public class BooleanChromosom extends Chromosom implements Cloneable, Serializable
{
	/**
	 * Der Inhalt des Chromosoms (quasi die DNS).
	 */
	private boolean[] field;

	/* (non-Javadoc)
	 * @see ga.IChromosom#mutate()
	 */
	public void mutate(double mutationRate)
	{
		// Anzahl der Muationsstellen errechnen.
		int mutationCount = (int) Math.round(mutationRate * field.length);

		int pos;
		for (long i = 0; i < mutationCount; ++i)
		{
			// Zufaellige eine Stelle in den Daten auswaehlen...
			pos = (int) Math.floor(Math.random() * field.length);
			field[pos] = !field[pos]; //...und veraendern.
		}
	}

	/* (non-Javadoc)
	 * @see ga.IChromosom#random()
	 */
	public void random()
	{
		// Feld mit zufaelligen Werten belegen
		for (int index = 0; index < field.length; ++index)
		{
			field[index] = ((int) Math.floor(Math.random() * 2) == 1);
		}
	}

	/**
	 * Legt ein boolisches Chromosom, mit den angegebenen Werten an.
	 * @param field Array von boolean Werten, mit denen das Chromosom initialisiert wird.
	 */
	public BooleanChromosom(boolean[] field)
	{
		// Feld kopieren
		this.field = (boolean[]) field.clone();
	}

	/**
	 * Legt ein boolisches, leeres Chromosom an.
	 */
	public BooleanChromosom()
	{
		this.field = new boolean[0];
	}

	/**
	 * Legt ein boolsches Chromosom, mit der angegebenen L&auml;nge an.
	 * Der Inhalt ist dabei unbestimmt.
	 * @param size L&auml;nge des Chromosoms.
	 */
	public BooleanChromosom(int size)
	{
		// Feld anlegen
		this.field = new boolean[size];
	}

	/**
	 * Liefert das boolsche Datenfeld des Chromosoms.
	 * @return Eine Kopie des Datenfelds.
	 */
	public boolean[] getField()
	{
		return (boolean[]) field.clone();
	}

	/**
	 * Setzt das Datenfeld des Chromosoms.
	 * @param field Das neue Datenfeld.
	 */
	public void setField(boolean[] field)
	{
		// Feld kopieren
		this.field = (boolean[]) field.clone();
	}

	/**
	 * Erstellt eine Kopie des Chromosoms.
	 */
	public Object clone()
	{
		// Kopie erstellen
		return new BooleanChromosom(field);
	}

	/**
	 * Gibt das Datenfeld als String aus.
	 */
	public String toString()
	{
		String result = "";
		for (int i = 0; i < field.length; ++i)
			if (field[i])
				result = result + "1";
			else
				result = result + "0";

		return result;
	}

    /* (non-Javadoc)
     * @see ga.Chromosom#copy(ga.Chromosom)
     */
    public void copy(Chromosom other) throws ClassCastException
    {
        BooleanChromosom o = (BooleanChromosom) other;
        System.arraycopy(o.field, 0, this.field, 0, this.field.length);
    }

    /* (non-Javadoc)
     * @see ga.Chromosom#crossover(ga.Chromosom, double, ga.Chromosom[])
     */
    public void crossover(Chromosom other, double twoPointCrossoverRate, Chromosom[] childChromosom) throws ClassCastException
    {
    	// Hier wuerde bei Fehler die ClassCastException geworfen.
		BooleanChromosom o = (BooleanChromosom) other;

		int l = this.field.length;

		double r;
		int pos1, pos2, tmp1, tmp2;

		// Kinderfeld uebernehmen
		// Hier wuerde bei Fehler die ClassCastException geworfen.
		BooleanChromosom[] children = (BooleanChromosom[]) childChromosom;

		r = Math.random(); // Ermitteln ob 1-Punkt oder 2-Punkt crossover

		if (r <= twoPointCrossoverRate)
		{
			// Zwei-Punkt-Crossover

			// Die zwei Schnittpunkte auswuerfeln
			tmp1 = (int) Math.floor(Math.random() * l);
			tmp2 = (int) Math.floor(Math.random() * l);
			pos1 = Math.min(tmp1, tmp2);
			pos2 = Math.max(tmp1, tmp2);

			// Chromosomsegmente tauschen
			System.arraycopy(this.field, 0, children[0].field, 0, pos1);
			System.arraycopy(o.field, 0, children[1].field, 0, pos1);
			System.arraycopy(
				this.field,
				pos1,
				children[1].field,
				pos1,
				pos2 - pos1);
			System.arraycopy(
				o.field,
				pos1,
				children[0].field,
				pos1,
				pos2 - pos1);
			System.arraycopy(
				this.field,
				pos2,
				children[0].field,
				pos2,
				this.field.length - pos2);
			System.arraycopy(
				o.field,
				pos2,
				children[1].field,
				pos2,
				o.field.length - pos2);
		}
		else
		{
			// Ein-Punkt-Crossover

			// Den Schnittpunkt auswuerfeln
			pos1 = (int) Math.floor(Math.random() * l);

			// Chromosomsegemente tauschen
			System.arraycopy(this.field, 0, children[1].field, 0, pos1);
			System.arraycopy(o.field, 0, children[0].field, 0, pos1);
			System.arraycopy(
				this.field,
				pos1,
				children[0].field,
				pos1,
				this.field.length - pos1);
			System.arraycopy(
				o.field,
				pos1,
				children[1].field,
				pos1,
				o.field.length - pos1);
		}
    }
}