/*
 * Copyright (C) 2004  Martin Hedler
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Last change:
 * 2004-02-17 by Martin Hedler (matin@hpfsc.de) 
 * 
 */
package ga;

import java.io.Serializable;
/**
 * Dient als Grundlage f&uuml;r spezialisierte Chromosomen. Ein Chromosom ist ein Tr&auml;ger f&uuml;r
 * Informationen. D.h. je nach Anwendung kann eine andere Kodierung gew&auml;hlt werden.
 * Jedoch muss dabei selbst der Mutations- und Crossoveralgorithmus implementiert werden.
 * Es liegen jedoch bereits spezialisierte Chromosomen f&uuml;r entsprechende Datentypen vor:
 * {@link BooleanChromosom} und {@link DoubleChromosom}.
 * 
 * @author Martin Hedler
 * @version 0.7
 */
public abstract class Chromosom implements Cloneable, Serializable
{
    /**
     * Mutiert den Inhalt des Chromosoms.
     * @param mutationRate Mutationswahrscheinlichkeit mit der ein Gen mutiert wird, bzw.
     * die Rate, die angibt, wieviel Prozent des Chromosoms mutiert wird.
     */
    public abstract void mutate(double mutationRate);

    /**
     * F&uuml;hrt ein Crossover zwischen diesem und einem anderen Chromosomen durch.
     * Es sind zwei Arten von Crossovern m&ouml;glich: 1-Punkt-Crossover und 2-Punkt-Crossover.
     * Die Wahrsheinlichkeit f&uuml;r ein 2-Punkt-Crossover wird angegeben. Daraus ergibt sich die
     * Wahrscheinlichkeit f&uuml;r ein 1-Punk-Crossover folgendermassen: P(1-Punkt) = 1 - P(2-Punkt).
     * @param other Das andere Chromosom mit dem dieses Chromosom gekreutzt wird.
     * @param twoPointCrossoverRate Wahrscheinlichkeit eines 2-Punkt-Crossovers.
     * @param childChromosom Ein Array mit der L&auml;ge 2, welches die beiden Ergebnischromosomen aufnehmen soll.
     * Die beiden Chromosome des Arrays m&uuml;ssen bereits existieren, so dass der Inhalt nur noch
     * hineinkopiert werden muss.
     * @throws ClassCastException Wird geworfen, wenn die beiden Chromosome nicht miteinander gekreutzt
     * werden koennen, z.B. weil sie von unterschiedlichen Typen sind. Oder wenn die Ergebnischromosome von einem
     * anderen Typ sind.
     */
    public abstract void crossover(
        Chromosom other,
        double twoPointCrossoverRate,
        Chromosom[] childChromosom)
        throws ClassCastException;

    /**
     * Ver&auml;ndert den Inhalt des Chromosoms zuf&auml;llig.
     */
    public abstract void random();

    /**
     * Erstellt eine Kopie des Chromosoms.
     * @return Eine Kopie des Chromosoms.
     */
    public Object clone() throws CloneNotSupportedException
    {
        return super.clone();
    }

    /**
     * Kopiert den Inhalt des angegebenen Chromosoms in sich selbst.
     * Es wird eine vollst&auml;ndige Kopie angelegt und nicht nur eine
     * Referenz auf die Attribute von <code>other</code>.
     * @param other Das Quellchromosom, von welchem die Werte &uuml;bernommen werden
     * sollen.
     */
    public abstract void copy(Chromosom other) throws ClassCastException;

}