/*
 * Copyright (C) 2004  Martin Hedler
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Last change:
 * 2004-02-17 by Martin Hedler (matin@hpfsc.de) 
 * 
 */
package ga;

import java.io.Serializable;

/**
 * Stellt ein Chromosom zur Verwaltung von double Werten zur Verf&uuml;gung.
 * @author Martin Hedler
 * @version 0.7
 */
public class DoubleChromosom extends Chromosom implements Cloneable, Serializable
{
	/**
	 * Der Inhalt des Chromosoms (quasi die DNS).
	 */
	private double[] field;
	
	// Mutationsrate, die auch beim Crossover gebraucht wird.
	private double mutationRate;
	
	// Wird fuer Crossover benoetigt.
	private double alpha;
	
	// Geben den Bereich an, in den sich die Zahlen bewegen sollen.
	private double minValue, maxValue;

	private double[] tmpElements;

    /**
	 * Legt ein double Chromosom, mit der angegebenen L&auml;nge an.
	 * Der Inhalt ist dabei unbestimmt.
	 * @param size L&auml;nge des Chromosoms.
	 */
	public DoubleChromosom(int size)
	{
		// Feld anlegen
		field = new double[size];

		// Defaultwerte setzen
		mutationRate = 0.15;
		alpha = 0.6;
		minValue = -1;
		maxValue = 1;
        tmpElements = new double[2];
	}

	/**
	 * Legt ein leeres double Chromosom an.
	 */
	public DoubleChromosom()
	{
		// Defaultwerte setzen
		this.field = new double[0];
		mutationRate = 0.15;
		alpha = 0.6;
		minValue = -1;
		maxValue = 1;
		tmpElements = new double[2];
	}

	/**
	 * Legt ein double Chromosom, mit den angegebenen Werten an.
	 * @param field Array von double Werten, mit denen das Chromosom initialisiert wird.
	 */
	public DoubleChromosom(double field[])
	{
		// Feld kopieren
		this.field = (double[]) field.clone();

		// Defaultwerte setzen
		mutationRate = 0.15;
		alpha = 0.6;
		minValue = -1;
		maxValue = 1;
        tmpElements = new double[2];
	}

	/**
	 * Setzt den Wertebereich, den die Zahlen in dem Chromosom einnehmen k&ouml;nnen.
	 * Die voreingestellten Werte sind -1 und +1.
	 * @param minValue Die untere Schranke.
	 * @param maxValue Die obere Schranke.
	 */
	public void setRange(double minValue, double maxValue)
	{
		this.minValue = Math.min(minValue, maxValue);
		this.maxValue = Math.max(minValue, maxValue);
	}

	/* (non-Javadoc)
	 * @see ga.Chromosom#mutate(double)
	 */
	public void mutate(double mutationRate)
	{
		// Anzahl der Mutationsstellen errechnen
		int mutationCount = (int) Math.floor(mutationRate * field.length * 64);

		long pos;
		for (long i = 0; i < mutationCount; ++i)
		{
			// Zufaellig eine Stelle in den Daten auswaehlen...
			pos = (int) Math.floor(Math.random() * field.length * 64);
			field[(int)(pos / 64)] = Double.longBitsToDouble((1 << (pos % 64)) ^ Double.doubleToLongBits(field[(int)(pos / 64)]));
		}

		this.mutationRate = mutationRate;
	}

	/**
	 * F&uuml;hrt ein Crossover von zwei double Werten durch. Das Ergebnis wird in das
	 * Array <code>dest</code> von double Werten mit der L&uml;nge 2 kopiert. Dieses Array
	 * enth&auml;lt die Kinderwerte.
	 * @param p1 Elternwert.
	 * @param p2 Elternwert.
	 * @param dest Array mit der L&auml;nge 2, welches die Kinderwerte aufnimmt. Es muss bereits
	 * angelegt wurden sein.
	 */
	private void crossoverElements(double p1, double p2, double[] dest)
	{
		double r, korr, e1, e2;

		r = Math.random();
//		alpha = Math.random();

		if (p1 != 0 | p2 != 0)
		{
			korr = (p1 * p2) / (Math.abs(p1) + Math.abs(p2));
		}
		else
		{
			korr = 0;
		}

		korr = 2 * mutationRate * r * korr;

		dest[0] = alpha * p1 + (1 - alpha) * p2 + korr;
		dest[1] = (1 - alpha) * p1 + alpha * p2 + korr;
	}

	/* (non-Javadoc)
	 * @see ga.Chromosom#random()
	 */
	public void random()
	{
		// Feld mit zufaelligen Werten belegen
		for (int index = 0; index < field.length; ++index)
		{
			field[index] = Math.random() * (maxValue - minValue) + minValue;
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException
	{
		// Kopie erstellen
		DoubleChromosom d = (DoubleChromosom) super.clone();
		d.field = new double[this.field.length];
		for (int index = 0; index < this.field.length; ++index)
		{
			d.field[index] = this.field[index];
		}
		d.mutationRate = mutationRate;
		d.alpha = alpha;
		d.minValue = minValue;
		d.maxValue = maxValue;
		return d;
	}

	/**
	 * Liefert das double Datenfeld des Chromosoms.
	 * @return Eine Kopie des Datenfelds.
	 */
	public double[] getField()
	{
		return field;
//		return (double[]) field.clone();
	}

	/**
	 * Gibt das Datenfeld als String aus.
	 */
	public String toString()
	{
		String result = new String();

		for (int index = 0; index < field.length; ++index)
		{
			result = result + String.valueOf(field[index]);

			if (index + 1 < field.length)
			{
				result = result + ", ";
			}
		}

		return result;
	}

    /* (non-Javadoc)
     * @see ga.Chromosom#crossover(ga.Chromosom, double, ga.Chromosom[])
     */
    public void crossover(Chromosom other, double twoPointCrossoverRate, Chromosom[] childChromosom) throws ClassCastException
    {
		// Hier wird bei Fehler die ClassCastException geworfen.
		DoubleChromosom o = (DoubleChromosom) other;

		int l = this.field.length;

		double r;
		int pos1, pos2, tmp1, tmp2;

		// Kinder uebernehmen
		// Hier wird bei Fehler die ClassCastException geworfen.
		r = Math.random(); // Ermitteln ob 1-Punkt oder 2-Punkt crossover

		if (r <= twoPointCrossoverRate)
		{
			// Zwei-Punkt-Crossover

			// Die zwei Schnittpunkte auswuerfeln
			tmp1 = (int) Math.floor(Math.random() * l);
			tmp2 = (int) Math.floor(Math.random() * l);
			pos1 = Math.min(tmp1, tmp2);
			pos2 = Math.max(tmp1, tmp2);

			// Chromosomsegmente tauschen
			System.arraycopy(this.field, 0, ((DoubleChromosom) childChromosom[0]).field, 0, pos1);
			System.arraycopy(o.field, 0, ((DoubleChromosom) childChromosom[1]).field, 0, pos1);

			crossoverElements(this.field[pos1], o.field[pos1], tmpElements);
			((DoubleChromosom) childChromosom[0]).field[pos1] = tmpElements[0];
			((DoubleChromosom) childChromosom[1]).field[pos1] = tmpElements[1];

			if (pos2 - pos1 > 1)
			{
				System.arraycopy(
					this.field,
					pos1 + 1,
					((DoubleChromosom) childChromosom[1]).field,
					pos1 + 1,
					pos2 - pos1 - 1);
				System.arraycopy(
					o.field,
					pos1 + 1,
					((DoubleChromosom) childChromosom[0]).field,
					pos1 + 1,
					pos2 - pos1 - 1);
			}

			crossoverElements(this.field[pos2], o.field[pos2], tmpElements);
			((DoubleChromosom) childChromosom[0]).field[pos1] = tmpElements[1];
			((DoubleChromosom) childChromosom[1]).field[pos1] = tmpElements[0];

			if (this.field.length - pos2 > 1)
			{
				System.arraycopy(
					this.field,
					pos2 + 1,
					((DoubleChromosom) childChromosom[0]).field,
					pos2 + 1,
					this.field.length - pos2 - 1);
			}

			if (o.field.length - pos2 > 1)
			{
				System.arraycopy(
					o.field,
					pos2 + 1,
					((DoubleChromosom) childChromosom[0]).field,
					pos2 + 1,
					o.field.length - pos2 - 1);
			}
		}
		else
		{
			// Ein-Punkt-Crossover

			// Den Schnittpunkt auswuerfeln
			pos1 = (int) Math.floor(Math.random() * l);

			// Chromosomsegemente tauschen
			System.arraycopy(this.field, 0, ((DoubleChromosom) childChromosom[1]).field, 0, pos1);
			System.arraycopy(o.field, 0, ((DoubleChromosom) childChromosom[0]).field, 0, pos1);

			crossoverElements(this.field[pos1], o.field[pos1], tmpElements);
			((DoubleChromosom) childChromosom[0]).field[pos1] = tmpElements[0];
			((DoubleChromosom) childChromosom[1]).field[pos1] = tmpElements[1];

			if (this.field.length - pos1 > 1)
			{
				System.arraycopy(
					this.field,
					pos1,
				((DoubleChromosom) childChromosom[0]).field,
					pos1,
					this.field.length - pos1 - 1);
			}

			if (o.field.length - pos1 > 1)
			{
				System.arraycopy(
					o.field,
					pos1,
				((DoubleChromosom) childChromosom[0]).field,
					pos1,
					o.field.length - pos1 - 1);
			}
		}
    }

    /* (non-Javadoc)
     * @see ga.Chromosom#copy(ga.Chromosom)
     */
    public void copy(Chromosom other) throws ClassCastException
    {
        // Hier wird bei Fehler die ClassCastException geworfen.
        DoubleChromosom o = (DoubleChromosom) other;
        
        this.alpha = o.alpha;
        this.maxValue = o.maxValue;
        this.minValue = o.minValue;
        this.mutationRate = o.mutationRate;
        System.arraycopy(o.field, 0, this.field, 0, this.field.length);
    }

    /**
     * @param ds
     */
    public void setField(double[] ds)
    {
        field = ds;
    }

}
