/*
 * Copyright (C) 2004  Martin Hedler
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Last change:
 * 2004-02-17 by Martin Hedler (matin@hpfsc.de) 
 * 
 */
package ga;

import java.io.Serializable;
/**
 * Ein Individuum verwaltet mehrere Chromosome. Es ist definiert durch einen sogenannten Chromosomensatz,
 * der die Merkmale des Individuums enth&auml;lt. Ein Individuum besitz ausserdem noch
 * eine Fitnes, die die Qualit&auml;t des Individuums beschreibt.
 * 
 * @version 0.7
 * @author Martin Hedler
 */
public class Individual implements Comparable, Cloneable, Serializable
{
    /**
     * Die aktuelle Fitness des Individuums.
     */
    private double fitnes;

    /**
     * Der Chromosomensatz des Individuums.
     */
    private Chromosom[] chromosomSet;

	/*
	 * Wird alles drei zum Chrossover benoetigt.
	 */
	private Chromosom[] chromosomSetChild1;
	private Chromosom[] chromosomSetChild2;
	private Chromosom[] tmpChromosom;

    /**
     * Legt ein neues Individuum mit dem angegebenen Chromosomensatzes an.
     * @param chromosomSet Der Chromosomensatz des Individuums.
     */
    public Individual(Chromosom[] chromosomSet)
    {
        // Fitness ist 0
        fitnes = 0;
        // Chromosomensatz anlegen
        this.chromosomSet = chromosomSet;
        
        // Temporaere Variablen initialisieren
		tmpChromosom = new Chromosom[2];
    }

    /**
     * Ver&auml;ndert den Chromosomensatz des Individuums zuf&auml;llig.
     */
    public void random()
    {
        for (int index = 0; index < chromosomSet.length; ++index)
        {
            chromosomSet[index].random();
        }
    }

    /**
     * Erstellt eine Kopie des Individuums, mit gleichem Chromosomensatz und gleicher
     * Fitnes.
     */
    public Object clone() throws CloneNotSupportedException
    {
        Individual result = (Individual) super.clone();
        result.chromosomSet = new Chromosom[this.chromosomSet.length];
        for (int index = 0; index < this.chromosomSet.length; ++index)
        {
			result.chromosomSet[index] =
				(Chromosom) this.chromosomSet[index].clone();

        }
        return result;
    }

	/*
	 *  (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
    public int compareTo(Object o)
    {
		return (this.fitnes > ((Individual) o).fitnes)? 1:this.fitnes < ((Individual) o).fitnes ? -1:0;
    }

    /**
     * Gibt die aktuelle Fitness des Individuums zurueck. Je gr&ouml;sser die Fitnes
     * ist, umso besser ist das Individuum.
     * @return Die Fitnes des Individuums.
     */
    public double getFitnes()
    {
        return fitnes;
    }

    /**
     * Setze die aktuelle Fitness des Individuums. Je gr&ouml;sser die Fitnes
     * ist, umso besser ist das Individuum.
     * @param d Die neue Fitness.
     */
    public void setFitnes(double d)
    {
        fitnes = d;
    }

    /**
     * Mutiert den Chomosomensatz des Individuums.
     * @param mutationRate Wahrscheinlichkeit, dass ein Merkmal ver&auml;ndert wird.
     */
    public void mutate(double mutationRate)
    {
        for (int index = 0; index < chromosomSet.length; ++index)
        {
            chromosomSet[index].mutate(mutationRate);
        }
    }

	/**
	 * Kreuze diese Individuum mit einem anderen. Dabei muss die Wahrscheinlichkeit eines 2-Punkt-Crossovers
	 * mit angegeben werden und das Array, in welches die Kinder kopiert werden sollen.
	 * Das Array muss bereits zwei Individuuen enthalten, da diese nur neu belegt
	 * und nicht neu erstellt werden. 
	 * @param other Das andere Individuum mit dem dieses gekreutzt werden soll.
	 * @param twoPointCrossoverRate Die Wahrscheinlichkeitsrate eines 2-Punkt-Crossovers.
	 * @param children Ein Array mit der L&auml;nge 2, welches die Kinder enth&auml;lt, welche bei der Kreuzung
	 * enststanden sind. Die beiden Kinder (Individuen) m&uuml;ssen bereits existieren, da ihre Chromosomen
	 * nur ver&auml;ndert werden. 
	 */
	public void crossover(
		Individual other,
		double twoPointCrossoverRate, Individual[] children)
	{
		// Chromosomensaetze der beiden Kinder erzeugen
		chromosomSetChild1 = children[0].chromosomSet;
		chromosomSetChild2 = children[1].chromosomSet;

		for (int index = 0; index < chromosomSet.length; ++index)
		{
			// Jeweils passende Chromosomen kreuzen
			tmpChromosom[0] = chromosomSetChild1[index];
			tmpChromosom[1] = chromosomSetChild2[index];

			chromosomSet[index].crossover(
					other.chromosomSet[index],
					twoPointCrossoverRate, tmpChromosom);
		}
	}


    /**
     * Liefert das Chromosom zur&uuml;ck, welches sich an der von index angegebenen
     * Position befindet.
     * @param index Die Position des gew&uuml;nschten Chromosoms.
     * @return Das Chromosom an der Potition index.
     * @throws IndexOutOfBoundsException Wird geworfen, wenn index &lt; 0 oder
     * index &gt;= getChromosomCount() ist.
     */
    public Chromosom getChromosom(int index) throws IndexOutOfBoundsException
    {
        return this.chromosomSet[index];
    }

    /**
     * Liefert die Anzahl der Chromosomen des Individuums.
     * @return Chromosomenanzahl.
     */
    public int getChromosomCount()
    {
        return this.chromosomSet.length;
    }

    public void setChromosomSet(Chromosom[] chromosomSet)
    {
        // Chromosomensatz anlegen
        this.chromosomSet = chromosomSet;
    }
    
    public Chromosom[] getChromosomSet()
    {
    	return this.chromosomSet;
    }

    public Individual()
    {
    	this(null);
    }
}