/*
 * Copyright (C) 2004  Martin Hedler
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Last change:
 * 2004-02-17 by Martin Hedler (matin@hpfsc.de) 
 * 
 */
package ga;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;

/**
 * Verwaltet eine Menge an Individuen und f&uuml;hrt den
 * Vortpflanzungs und Selektionsprozess durch.
 * @author Martin Hedler
 */
public abstract class Population implements Serializable
{
	/** Liste der Individuen in der Population. */
    private Individual memory[];
    
    /** Anzahl der momentan lebenden Individuen. */
    private int liveCount;
    
    /** Vektor zur Zwischenspeicherung von Individuen. */
	private Individual tempVector[];
	
	/** Vektor zum Speichern der Eltern. */
	private Individual parentVector[];
	
	/** Vektor zum Speichern der Kinder. */
	private Individual childVector[];
	
	/** Comperator zur Sortierung der Individuen. */
	private NegComparator negCom;
	
    /** Anzahl der Individuen in der Population. */
    private int individualCount;

    /** Anzahl der Kinder die in einer Generation zur Welt kommen. */
    private int childrenCount;

    /** Gr&ouml;sse einer Gruppe bei der Paarung. */
    private int groupSize;

	/** Wahrscheinlichkeit eines Crossovers. */
	private double crossoverRate;

    /** Wahrscheinlichkeitsrate f&uuml;r ein 2-Punkt-Crossover. */
    private double twoPointCrossoverRate;

    /** Mutationswahrsceinlichkeit. */
    private double mutationRate;

    /** Gibt an, ob die Population bereits initialisiert wurde. */
    private boolean notInitialized;

	/**
     * Erstellt eine neue Population mit Individuen vom Typ type.
     * @param type Urindividuum, welches als Muster f&uuml;r alle anderen Individuuen dient.
     * @param individualCount Anzahl der Individuen in einer Generation, die erzeugt werden sollen.
	 * @param childrenCount Anzahl der Kinder, die in jeder Generation erzeugt werden sollen.
	 * @param groupSize Gr&ouml;sse der Gruppen, die zur Paarung gebildet werden.
	 * @param crossoverRate Wahrscheinlichkeit dass ein Crossovers stattfindet.
	 * @param twoPointCrossoverRate Wahrscheinlichkeit fuer 2-Punkt-Crossover, falls &uuml;berhaupt
	 * ein Crossover stattfinden sollte.
	 * @param mutationRate Muationswahrscheinlichkeit.
	 */
    public Population(
        Individual type,
        int individualCount,
        int childrenCount,
        int groupSize,
        double crossoverRate,
        double twoPointCrossoverRate,
        double mutationRate)
    {
        // Speicher anlegen, der eine ganze Population aufnehmen kann.
        this.memory = new Individual[individualCount + childrenCount];
		// Speicher anlegen fuer den in sex() genutzen parentVector
		this.parentVector = new Individual[2];
		// Speicher anlegen fuer den in sex() genutzen tempVector
		this.tempVector = new Individual[individualCount];
		// Kindervecktor anlegen (ebenfalls in sex() genutzt)
		this.childVector = new Individual[2];

        // Individuenanzahl speichern
        this.individualCount = individualCount;
        // Zahl der momentan lebenden Individuuen speichern
        this.liveCount = individualCount;
        // Anzahl der Kinder je Generation speichern
        this.childrenCount = childrenCount;
        // Groesse einer Gruppe bei der Paarung
        this.groupSize = groupSize;
        // Wahrscheinlichkeit eines Crossovers speichern.
        this.crossoverRate = crossoverRate;
        // 2-Punkt-Crossover Wahrscheinlichkeit
        this.twoPointCrossoverRate = twoPointCrossoverRate;
        // Mutationsrate
        this.mutationRate = mutationRate;

		// Speichern, dass Population noch nicht initialisiert wurde.
        this.notInitialized = true;

        try
        {
            for (int i = 0; i < individualCount + childrenCount; ++i)
            {
                // Neues Individuen nach Muster erstellen
                memory[i] = (Individual) type.clone(); // TODO Eventuell durch Contructoraufruf ersetzen.
                // Genetische Information zufaellig veraendern
                memory[i].random();
            }
        }
        catch (CloneNotSupportedException e)
        {
            System.out.println("Kopiervorgang gescheitert");
        }
        
        // Einen Comparator zur Sortierung anlegen
        this.negCom = new NegComparator();
    }

	/**
	 * Erstellt eine neue Population mit Individuen vom Typ type. Mit Voreinstellungen
	 * f&uuml;r:<br>
	 * <code>childrenCount = 2</code><br>
	 * <code>groupSize = 5</code><br>
	 * <code>crossoverRate = 0.7</code><br>
	 * <code>twoPointCrossoverRate = 0.4</code><br>
	 * <code>muationRate = 0.02</code><br>
	 * 
	 * @param type Urindividuum, welches als Muster f&uuml;r alle anderen Individuuen dient.
	 * @param count Anzahl der Individuen in einer Generation, die erzeugt werden sollen.
	 */
	public Population(Individual type, int count)
	{
		this(type, count, 2, 5, 0.7, 0.4, 0.02);
	}

    /**
     * Erzeugt neue Individuen und f&uuml;hgt diese in die Population ein.
     */
    protected void sex()
    {
        int groupSize; // Groesse einer Gruppe
        int index;
        int tempVectorSize = tempVector.length;

		// Lebende Individuuen in tmpVector kopieren
		System.arraycopy(this.memory, 0, this.tempVector, 0, this.tempVector.length);

		// Groesstmoegliche Gruppengroesse ermitteln
        groupSize = individualCount / childrenCount;

        groupSize = Math.min(groupSize, this.groupSize);

        // Jede Gruppe durchgehen
        for (int groupIndex = 0; groupIndex < childrenCount; ++groupIndex)
        {
            // Fuer jede Gruppe groupSize Individuen suchen
            for (int round = 0; round < groupSize; ++round)
            {
                // Individuen zufaellig ermitteln
                index = (int) Math.floor(Math.random() * tempVectorSize);

                // Wenn noch kein Individuum in der Gruppe enthalten ist...
                if (round == 0)
                {
                    // Erstes Individuum in Gruppe einfuehgen
                    parentVector[groupIndex & 1] = tempVector[index];
                }
                else
                    // Ansonsten: Wenn das neugefundene Individuum besser ist, als das alte...
                    if (tempVector[index].getFitnes()
                        > parentVector[groupIndex & 1].getFitnes())
                    {
                        // ...altes Individuum durch neues ersetzen.
                        parentVector[groupIndex & 1] = tempVector[index];
                    }

                // Gefundene Individuum aus Liste loeschen, damit es nicht noch einmal gefunden wird
                --tempVectorSize;
                // Individuum wird ueberschrieben und Liste wird verkleinert
                tempVector[index] = tempVector[tempVectorSize];
            }

            // Wenn zwei Eltern ermittelt wurden... (Ungerade Gruppe)
            if ((groupIndex & 1) == 1)
            {
                // Individuuen paaren und Kinder speichern
                childVector[0] = this.memory[this.liveCount];
                childVector[1] = this.memory[this.liveCount + 1];
                
                // Crossover durchf&uuml;hren, wenn der Zufall es will
                if (Math.random() < this.crossoverRate)
                {
	                parentVector[0].crossover(parentVector[1], twoPointCrossoverRate, childVector);
                }

                // Kinder mutieren
                childVector[0].mutate(mutationRate);
                childVector[1].mutate(mutationRate);

                // Populationsgroesse aktualisieren
                this.liveCount += 2;
            }
        }
    }

    /**
     * Liefert das beste Individuum einer Generation zur&uuml;ck.
     * @return Das beste Individuum.
     */
    public Individual getBest()
    {
        return this.memory[0];
    }

    /**
     * Schlechte Individuen aus der Population entfernen.
     */
    protected void death()
    {
		// Soviele Kinder wie dazugekommen sind...
		// ... schlechte Individuuen aus Population loeschen.
    	this.liveCount -= this.childrenCount;
    }

    /**
     * F&uuml;hrt <code>count</code> Evolutionsschritte durch. In einem Schritt werden:
     * 1. neue Individuen erzeugt (Paarung / Kinder)<br>
     * 2. die Generation bewertet<br>
     * 3. und die schlechstesten Individuen geloescht.
     * @param count Anzahl der durchzuf&uuml;hrenden Schritte.
     */
    public void evolutionStep(int count)
    {
        if (this.notInitialized)
        {
            // Population bewerten.
            // Nur die lebenden Individuen bewerten
            System.arraycopy(memory, 0, tempVector, 0, liveCount);
            calculate(tempVector);
            // Individuen nach Fitnes sortieren (das beste ist an Pos. 0)
            Arrays.sort(memory, this.negCom);
        }

        for (int step = 0; step < count; ++step)
        {
            // Kinder zeugen
            sex();
            // Fitness der ganzen Generation berechnen
            calculate(memory);
            // Population sortieren
            Arrays.sort(memory, this.negCom);

            // Schlechte Individuen aus Population entfernen
            death();
        }
    }

    /**
     * Berechnet die Fitness der Individuen der Population.
     * @param l Liste der Individuen.
     */
    protected abstract void calculate(Individual l[]);

    /**
     * Liefert die Anzahl der Kinder zur&uuml;ck, die in jeder Generation geboren werden.
     * @return Anzahl der je Generation erzeugten Kinder.
     */
    public int getChildrenCount()
    {
        return childrenCount;
    }

    /**
     * Liefert die Gr&ouml;sse einer Paaraungsgruppe zur&uuml;ck.
     * @return Gr&ouml;sse einer Gruppe.
     */
    public int getGroupSize()
    {
        return groupSize;
    }

    /**
     * Zum Setzen der Anzahl der Kinder, die in jeder Generation geboren werden sollen.
     * @param s Anzahl der Kinder. Ist diese Zahl ungerade, so wird auf die n&auml;chst h&ouml;here
     * Zahl erh&ouml;ht.
     */
    public void setChildrenCount(int s)
    {
        if (s % 2 == 1)
        {
            ++s;
        }
     
     	if (s > this.individualCount)
     	{
			throw new IllegalArgumentException();
     	}
        
        this.childrenCount = s;
        
		// Neuen Speicher anlegen, der eine ganze Population aufnehmen kann.
		Individual tmp[] = new Individual[individualCount + childrenCount];

		// Existierenden Speicher kopieren
		System.arraycopy(this.memory, 0, tmp, 0, Math.min(this.memory.length, tmp.length));

		// Die noch leeren Felder mit Individuen fuellen.
		try
		{
			for (int i = Math.min(this.memory.length, tmp.length); i < tmp.length; ++i)
			{
				// Neues Individuen nach Muster erstellen
				tmp[i] = (Individual) memory[0].clone(); // TODO Eventuell durch Contructoraufruf ersetzen.
				// Genetische Information zufaellig veraendern
				tmp[i].random();
			}
		}
		catch (CloneNotSupportedException e)
		{
			System.out.println("Kopiervorgang gescheitert");
		}
		
		// Neuen Speicherbereich einhaengen.
		this.memory = tmp;
    }

	/**
	 * Zum Setzen der Populationsgr&ouml;sse.
	 * @param s Anzahl der Individuen.
	 */
	public void setIndividualCount(int s)
	{
		if (s < this.childrenCount)
		{
			throw new IllegalArgumentException();
		}
        
		this.individualCount = s;

		// Neuen Speicher anlegen, der eine ganze Population aufnehmen kann.
		Individual tmp[] = new Individual[individualCount + childrenCount];

		// Existierenden Speicher kopieren
		System.arraycopy(this.memory, 0, tmp, 0, Math.min(this.memory.length, tmp.length));

		// Die noch leeren Felder mit Individuen fuellen.
		try
		{
			for (int i = Math.min(this.memory.length, tmp.length); i < tmp.length; ++i)
			{
				// Neues Individuen nach Muster erstellen
				tmp[i] = (Individual) memory[0].clone(); // TODO Eventuell durch Contructoraufruf ersetzen.
				// Genetische Information zufaellig veraendern
				tmp[i].random();
			}
		}
		catch (CloneNotSupportedException e)
		{
			System.out.println("Kopiervorgang gescheitert");
		}
		
		// Neuen Speicherbereich einhaengen.
		this.memory = tmp;
		// Speicher anlegen fuer den in sex() genutzen tempVector
		this.tempVector = new Individual[individualCount];
		// Zahl der momentan lebenden Individuuen speichern
		this.liveCount = individualCount;
		// Speichern, dass Population noch nicht initialisiert wurde.
		this.notInitialized = true;
	}

    /**
     * Setzt die Gr&ouml;sse einer Paaraungsgruppe.
     * @param s Gruppengr&ouml;sse.
     */
    public void setGroupSize(int s)
    {
        groupSize = s;
    }

    /**
     * Liefert die Summe der Fitnes aller zur Zeit lebenden Individuen zur&uuml;ck.
     * @return Populationsfitnes.
     */
    public double getPopulationFitnes()
    {
        double fitnes = 0;

        for (int i = 0; i < this.liveCount; ++i)
        {
            fitnes += memory[i].getFitnes();
        }

        return fitnes;
    }

    /**
     * Liefert die Wahrscheinlichkeit f&uuml;r ein 2-Punkt-Crossover zur&uuml;ck.
     * @return Wahrscheinlichkeit f&uuml;r ein 2-Punkt-Crossover.
     */
    public double getTwoPointCrossoverRate()
    {
        return twoPointCrossoverRate;
    }

    /**
     * Setzt die Wahrscheinlichkeit daf&uuml;r, dass es zu einem 2-Punkt-Crossover kommt.
     * @param twoPointCrossoverRate Wahrscheinlichkeit f&uuml;r ein 2-Punkt-Crossover.
     */
    public void setTwoPointCrossoverRate(double twoPointCrossoverRate)
    {
        this.twoPointCrossoverRate = twoPointCrossoverRate;
    }

    /**
     * Liefert die Muationsrate zur&uuml;ck.
     * @return Muationswahrscheinlichkeit.
     */
    public double getMutationRate()
    {
        return mutationRate;
    }

    /**
     * Setzt die Mutationswahrscheinlichkeit.
     * @param mutationRate Muatationsrate.
     */
    public void setMutationRate(double mutationRate)
    {
        this.mutationRate = mutationRate;
    }

    /**
     * Liefert das Individuum zur&uuml;ck, welches sich an der durch Index angegebenen
     * Position befindet.
     * @param index Die Position des gew&uuml;nschten Individuums.
     * @return Das Individuum an der Position index.
     * @throws IndexOutOfBoundsException Wird geworfen, wenn index &lt; 0 | index &gt>=
     * getIndividualCount() ist.
     */
    public Individual getIndividual(int index) throws IndexOutOfBoundsException
    {
        return this.memory[index];
    }

	/**
	 * F&uuml;gt an die durch <code>index</code> angegebenen Position ein Individuum
	 * ein.
	 * @param index Position in der Population an der das Individuum eingef&uuml;gt werden soll.
	 * @param i Das Individuum, welches eingef&uuml;gt werden soll.
	 * @throws IndexOutOfBoundsException Wird geworfen, wenn <code>index</code> gr&ouml;sser ist, als
	 * die Population (Anzahl der Individuuen) oder kleiner Null.
	 */
	public void setIndividual(int index, Individual i) throws IndexOutOfBoundsException
	{
		// TODO ueberarbeiten, wegen Speicherallokation und Freigabe (CopyFrom)
		this.memory[index] = i;
	}

    /**
     * Liefert die Anzahl der Individuen zur&uuml;ck, die sich in der Population
     * befinden.
     * @return Die anktuelle Anzahl der Individuen.
     */
    public int getIndividualCount()
    {
        return this.individualCount;
    }

    /**
     * Vernichtet alle Individuuen, bis aus die besten und fuellt die Population
     * mit neuen zufaellig erzeugten Individuuen auf.
     * @param aliveCount Anzahl der Individuen die am Leben bleiben sollen.
     */
    public void chaos(int aliveCount)
    {
        for (int i = aliveCount; i < this.liveCount; i++)
        {
            this.memory[i].random();
        }
    }
    
    /**
     * Liefert die Wahrscheinlichkeit eines Crossover.
     * @return Crossoverwahrscheinlichkeit.
     */
    public double getCrossoverRate()
    {
        return crossoverRate;
    }

    /**
     * Setzt die Wahrscheinlichkeit eines Crossovers.
     * @param d Crossoverwahrscheinlichkeit.
     */
    public void setCrossoverRate(double d)
    {
        crossoverRate = d;
    }

	/**
	 * Der NegComparator dreht die Sortierreihenfolge beim Vergleichen um, so
	 * das von gross nach klein sortiert wird. 
	 */
	private class NegComparator implements Comparator
	{
		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2)
		{
			return -((Comparable)o1).compareTo(o2);
		}
	}   
}